module Main where

import Data.Semigroup
--import Data.Monoid
import Data.Char

-- class Semigroup m where
--   (<>) :: m -> m -> m
-- associativo

-- class Monoid m where
--   mempty :: m 
--   mappend = (<>)
--   mconcat :: [m] -> m = fold 

-- Sum, Product, First, Last, Max, Min
xs :: [Int]
xs = [1..10]

-- >>> foldMap Sum xs
-- Sum {getSum = 55}

-- >>> foldMap Product xs
-- Product {getProduct = 3628800}

-- >>> foldMap Max xs
-- Max {getMax = 10}

-- >>> foldMap Min xs
-- Min {getMin = 1}

-- >>> foldMap (All . even) xs
-- All {getAll = False}

-- >>> foldMap (Any . even) xs
-- Any {getAny = True}

-- stimesMonoid
-- >>> stimesMonoid 10 "a"

-- fast fibonacci matrix
{- Podemos calcular o Fib(n) recuperando o elemento (0,1) da matriz:

|0 1|^n 
|1 0|
-}

data FibMtx = FibMtx { f00 :: Integer
                     , f01 :: Integer
                     , f10 :: Integer
                     , f11 :: Integer
                     } deriving Show

instance Semigroup FibMtx where
  FibMtx l00 l01 l10 l11 <> FibMtx r00 r01 r10 r11 =
    FibMtx (l00*r00 + l01*r10) (l00*r01 + l01*r11) (l10*r00+l11*r10) (l10*r01 + l11*r11)

instance Monoid FibMtx where
  mempty = FibMtx 1 0 0 1

fib :: Integer -> Integer
fib n = f01
      $ mtimesDefault n
      $ FibMtx 0 1 1 1

-- >>> map fib [1..12]
-- [1,1,2,3,5,8,13,21,34,55,89,144]

-- Functor

-- fmap id xs = xs
-- fmap f (fmap g xs) = fmap (f.g) xs
bool2int True = 1
bool2int False = 0

tarefas xs = fmap show $ fmap bool2int $ fmap odd $ fmap (+1) $ fmap length xs

tarefas' xs = fmap f xs
  where f = show . bool2int . odd . (+1) . length

data Uno a = Uno a deriving Show

instance Functor Uno where
   -- (a -> b) -> Uno a -> Uno b
   -- x :: a
   -- f x :: b
  fmap f (Uno x) = Uno (f x)

x = Uno 10 :: Uno Int
fu = Uno even :: Uno (Int -> Bool)

-- >>> fmap show x
-- Uno "10"

-- ( (Int->Bool) -> Bool ) -> Uno (Int -> Bool) -> Uno Bool 
-- >>> fmap (\h -> h 2) fu 
-- Uno True

instance Applicative Uno where
  pure x = Uno x
  Uno f <*> Uno x = Uno (f x)

-- >>> Uno even <*> Uno 2
-- Uno True

-- >>> Uno (+) <*> Uno 1 <*> Uno 2
-- Uno 3

-- Uno (+) <*> Uno 1 = Uno ((+) 1) = Uno (+1)
-- Uno (+1) <*> Uno 2 = Uno ( (+1) 2) = Uno 3

instance Monad Uno where
  -- >>= :: Uno a -> (a -> Uno b) -> Uno b
  Uno x >>= f = f x

-- >>> Uno 1 >>= \x -> Uno (2*x)
-- Uno 2

int2Uno x = Uno (even x)

fmu :: Uno Int -> Uno String
fmu u = do x <- u
           y <- int2Uno x
           if y
              then return "Ok"
              else return "Bogus"

-- >>> fmu (Uno 2)           
-- Uno "Ok"

type Semente = Integer
data Aleatorio a = Aleatorio (Semente -> (a, Semente))

instance Functor Aleatorio where
  -- (a -> b) -> Aleatorio a -> Aleatorio b
  -- (a -> b) -> (S -> (a, S)) -> (S -> (b, S))
  fmap f (Aleatorio g) = Aleatorio (\s -> let (a, s') = g s
                                          in  (f a, s'))

geraNum :: Semente -> (Integer, Semente)
geraNum semente = (semente `mod` 96 + 32, (semente*123456 + 123) `mod` 9531)

gerador = Aleatorio geraNum

executa :: Aleatorio a -> Semente -> (a, Semente)
executa (Aleatorio g) s = g s

-- >>> executa gerador 10
-- (32,43)

-- >>> executa (fmap (chr.fromInteger) gerador) 43
-- ('\131',175)

-- >>> executa (fmap even gerador) 43
-- (False,175)

geraSeq :: Int -> Aleatorio a -> Semente -> ([a], Semente)
geraSeq 0 _ s = ([], s)
geraSeq n aleat s = let (x, s')   = executa aleat s
                        (xs, s'') = geraSeq (n-1) aleat s'
                    in  (x:xs, s'')

-- >>> geraSeq 5 gerador 10                  
-- ([32,131,527,575,767],511)

instance Applicative Aleatorio where
  -- x :: a, Aleatorio a = Semente -> (a, Semente)
  pure x = Aleatorio (\s -> (x, s))

  -- Aleatorio (a -> b) <*> Aleatorio a ==> Aleatorio b
  -- S -> ((a->b), S)   <*> S -> (a, S) ==> S -> (b, S)
  Aleatorio f <*> Aleatorio x = Aleatorio (\s -> let (g, s1) = f s
                                                     (a, s2) = x s1
                                                 in  (g a, s2))

instance Monad Aleatorio where
  Aleatorio x >>= f = Aleatorio (\s -> let (g, s1) = x s
                                       in  executa (f g) s1)

geraStr :: Int -> Aleatorio String
geraStr 0 = return ""
geraStr n = do c  <- chr.fromInteger <$> gerador
               cs <- geraStr (n-1)
               return (c:cs)

geraStr' :: Int -> Aleatorio String
geraStr' n = sequence $ replicate n geraChar -- replicate me retorna [Aleatorio Char], mas eu quero Aleatorio [Char]
  where
    geraChar = fmap (chr.fromInteger) gerador

main :: IO ()
main = do
  putStrLn $ fst $ executa (geraStr 5) 13
  putStrLn $ fst $ executa (geraStr' 5) 13
  mapM_ putStrLn $ fst $ executa (traverse geraStr [1..10]) 10
